/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#include <vector>
#include <proj_api.h>
#include "ChartChooserDrawingArea.h"
#include "Exception.h"

ChartChooserDrawingArea::ChartChooserDrawingArea(
	BaseObjectType* base_object,
	const Glib::RefPtr<Gnome::Glade::Xml>& glade_xml) : Gtk::DrawingArea(base_object
	)
{
	Proj=pj_init_plus("+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs");
	if (!Proj) {
		throw Exception("Unable to initialize projection",Exception::LEVEL_ERROR);
	}

	XScale=1.0;
	XOfs=0.0;
	YScale=1.0;
	YOfs=0.0;
	ChartType=GeoChart::TYPE_SECTIONAL;
	Data=NULL;
	RouteData=NULL;

	add_events(Gdk::BUTTON_PRESS_MASK);
	add_events(Gdk::BUTTON_RELEASE_MASK);
	add_events(Gdk::POINTER_MOTION_MASK);
	/* Hint=don't send us another event until we've processed the last one
	 * and called get_pointer()
	 */
	add_events(Gdk::POINTER_MOTION_HINT_MASK);
}

ChartChooserDrawingArea::~ChartChooserDrawingArea()
{
	pj_free(Proj);
}

bool ChartChooserDrawingArea::on_expose_event(GdkEventExpose* ev) {

	if (BgImgPixbuf) {
		BgImgPixbuf->render_to_drawable(
			get_window(),
			get_style()->get_black_gc(),
			0,			/* src x offset */
			0,			/* src y offset */
			0,			/* dest x offset */
			0,			/* dest y offset */
			BgImgPixbuf->get_width(),
			BgImgPixbuf->get_height(),
			Gdk::RGB_DITHER_NONE,
			0,
			0);
	}

	DrawGreatCircleRoute(RouteData);

	Cairo::RefPtr<Cairo::Context> Context = get_window()->create_cairo_context();

	/* Set up brush */
	Context->set_source_rgba(1.0, 1.0, 1.0, 0.3);

	if (Data) {
		for (
			std::vector<GeoChart*>::iterator GeoCharti=Data->GetGeoCharts()->begin();
			GeoCharti!=Data->GetGeoCharts()->end();
			++GeoCharti
			) {
			if (
				(*GeoCharti)->GetGeoReferenced()==true &&
				(*GeoCharti)->GetType()==ChartType &&
				(GeoCharti!=*SelectedCharti) &&
				(GeoCharti!=HoveredCharti)
			) {
				DrawGeoChartBounds(Context,*GeoCharti);
			}
		}

		if (*SelectedCharti!=Data->GetGeoCharts()->end()) {
			Context->set_source_rgba(0.6, 1.0, 0.6, 0.8);
			DrawGeoChartBounds(Context,**SelectedCharti);
		}
	
		if (HoveredCharti!=Data->GetGeoCharts()->end()) {
			Context->set_source_rgba(0.0, 0.6, 1.0, 0.6);
			DrawGeoChartBounds(Context,*HoveredCharti);
		}
	}

	return true;
}

bool ChartChooserDrawingArea::on_button_release_event(GdkEventButton* event) {
	*SelectedCharti=GetChartByLoc(event->x,event->y);

	Redraw();
	return false;	/* Let the event fall through to the ChartChooserDialog */
}

bool ChartChooserDrawingArea::on_motion_notify_event(GdkEventMotion* event) {
	int xofs=0,yofs=0;
	std::vector<GeoChart*>::iterator CurrHoveri;

	/* This not only returns the mouse position but lets the event manager know that we're
	 * ready for another GdkEventMotion event whenever it's ready
	 */
	get_pointer(xofs,yofs);

	CurrHoveri=GetChartByLoc(xofs,yofs);

	if (CurrHoveri!=HoveredCharti) {
		HoveredCharti=CurrHoveri;
		Redraw();
	}

	return true;
}

void ChartChooserDrawingArea::Init(
	std::string BackgroundImgPath,
	DataStore *Data,
	enum Area_enum Area,
	std::vector<GeoChart*>::iterator *SelectedCharti,
	Route *RouteData
	) {
	this->BackgroundImgPath=BackgroundImgPath;
	this->Data=Data;
	this->HoveredCharti=HoveredCharti;
	this->SelectedCharti=SelectedCharti;
	this->HoveredCharti=Data->GetGeoCharts()->end();
	this->RouteData=RouteData;

	BgImgPixbuf = Gdk::Pixbuf::create_from_file(BackgroundImgPath.c_str());

	/* Size the drawing area to the size of the background image */
	set_size_request(BgImgPixbuf->get_width(),BgImgPixbuf->get_height());

	/* These were done crudely by hand and and have very low
	 * accuracy.  They are intended to be suitable for figuring out roughly what
	 * area a chart covers and not much else.
	 */
	switch (Area) {
		case AREA_CONTIGUOUS:
			SetScaleOfs(
				-81.7595556,24.5561111,380,223,	/* KEYW Key West Airport */
				-122.30931,47.44900,14,12		/* KSEA Seattle-Tacoma Airport */
			);
			break;
		case AREA_ALASKA:
			SetScaleOfs(
				-134.5762778,58.3549722,103,30,	/* PAJN Juneau Airport */
				-166.1100211,68.8751322,30,5	/* PALU Cape Lisburne Airport */
			);
			break;
		case AREA_HAWAII:
			SetScaleOfs(
				-155.04847016,19.7202630,37,17,	/* PHTO Hilo Airport */
				-159.7894444,22.0247222,4,1		/* PHBK Barking Sands Airport */
			);
			break;
	}
}

void ChartChooserDrawingArea::DrawGeoChartBounds(Cairo::RefPtr<Cairo::Context> Context,GeoChart *Chart) {
	double X1, Y1, X2, Y2;

	LonLatToPixel(Chart->GetWestBounding(),Chart->GetNorthBounding(),&X1,&Y1);
	LonLatToPixel(Chart->GetEastBounding(),Chart->GetSouthBounding(),&X2,&Y2);

	Context->rectangle(X1,Y1,fabs(X2-X1),fabs(Y2-Y1));
	Context->fill();
}

void ChartChooserDrawingArea::LonLatToPixel(double Lon, double Lat, double *X, double *Y) {
	projUV p;

	p.u=Lon*DEG_TO_RAD;
	p.v=Lat*DEG_TO_RAD;

	p=pj_fwd(p,Proj);

	*X=p.u*XScale+XOfs;
	*Y=p.v*YScale+YOfs;
}

/* Determine adjustments factors given 2 lon/lats and their pixel positions */
void ChartChooserDrawingArea::SetScaleOfs(double Lon1, double Lat1, double X1, double Y1,
	double Lon2, double Lat2, double X2, double Y2) {
	double XP1,YP1,XP2,YP2;

	/* Determinte the unadjusted values */
	XScale=1.0;
	XOfs=0;
	YScale=1.0;
	YOfs=0;

	LonLatToPixel(Lon1,Lat1,&XP1,&YP1);
	LonLatToPixel(Lon2,Lat2,&XP2,&YP2);

	/* Determine the required adjustments */
	XScale=(X1-X2)/(XP1-XP2);
	XOfs=X1-(XP1*XScale);
	YScale=(Y1-Y2)/(YP1-YP2);
	YOfs=Y1-(YP1*YScale);
}

void ChartChooserDrawingArea::Redraw() {
	Gdk::Rectangle rect(0,0,get_width(),get_height());
	get_window()->invalidate_rect(rect,1);
}

void ChartChooserDrawingArea::SetChartType(GeoChart::Type_enum ChartType) {
	this->ChartType=ChartType;
}

std::vector<GeoChart*>::iterator ChartChooserDrawingArea::GetChartByLoc(double X, double Y) {
	std::vector<GeoChart*>::iterator ChosenCharti;
	double X1,Y1,X2,Y2;

	if (Data) {
		ChosenCharti=Data->GetGeoCharts()->end();	/* Initialize to nothing valid being hovered over */
		for (std::vector<GeoChart*>::iterator GeoCharti=Data->GetGeoCharts()->begin();GeoCharti!=Data->GetGeoCharts()->end();++GeoCharti) {
			LonLatToPixel((*GeoCharti)->GetWestBounding(),(*GeoCharti)->GetNorthBounding(),&X1,&Y1);
			LonLatToPixel((*GeoCharti)->GetEastBounding(),(*GeoCharti)->GetSouthBounding(),&X2,&Y2);
			if (
				((*GeoCharti)->GetType()==ChartType) &&
				(X>=X1) &&
				(X<=X2) &&
				(Y>=Y1) &&
				(Y<=Y2)
			) {
				ChosenCharti=GeoCharti;
			}
		}
	} else {
		throw Exception("Invalid datastore",Exception::LEVEL_ERROR);
	}

	return ChosenCharti;
}

void ChartChooserDrawingArea::DrawGreatCircleRoute(Route *RouteData)
{
	double XPixel,YPixel;
	std::list<RoutePoint*> *LinRoutePoints;
	std::list<RoutePoint*>::iterator RoutePointi;

	if (!RouteData) {
		/* Nothing ot draw */
		return;
	}

	LinRoutePoints=RouteData->GetLinearRoutePoints();
	if (LinRoutePoints->size()<2) {
		/* Nothing to draw */
		return;
	}

	Cairo::RefPtr<Cairo::Context> Context = get_window()->create_cairo_context();

	/* Set up brush */
	Context->set_source_rgba(0.8, 0.2, 0.5, 0.4);
	Context->set_line_width(5.0);

	/* Go to each point */
	for (RoutePointi=LinRoutePoints->begin();RoutePointi!=LinRoutePoints->end();RoutePointi++)
	{
		LonLatToPixel((*RoutePointi)->GetLon(),(*RoutePointi)->GetLat(),&XPixel,&YPixel);

		if (RoutePointi==LinRoutePoints->begin()) {
			/* Go to the start position */
			Context->move_to(XPixel,YPixel);
		} else {
			/* Go to this position */
			Context->line_to(XPixel,YPixel);
		}
	}

	/* Draw the line */
	Context->stroke();
}



