/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#ifndef SHAPEFILE_H_
#define SHAPEFILE_H_

/* <cstdint> is currently experimental (according to Fedora 10), so use the C one */
#include <stdint.h>
#include <cstdio>
#include <vector>
#include <list>
#include <string>

#define SHAPEFILE_NODATA 1e-38
/* Version and File Code from the ESRI Shapefile Technical Description dated July 1998 */
#define SHAPEFILE_VERSION 1000
#define SHAPEFILE_FILECODE 9994

/* All integers are 32-bit signed bits.  _b suffix=Big Endian, _l suffix=Little Endian. */
/* All floats are 64-bit IEEE double-precisions floats, little endian. */

enum shapefile_shapetype_enum {
	SHAPETYPE_NULL,
	SHAPETYPE_POINT,
	SHAPETYPE_POLYLINE=3,
	SHAPETYPE_POLYGON=5,
	SHAPETYPE_MULTIPOINT=8,
	SHAPETYPE_POINTZ=11,
	SHAPETYPE_POLYLINEZ=13,
	SHAPETYPE_POLYGONZ=15,
	SHAPETYPE_MULTIPOINTZ=18,
	SHAPETYPE_POINTM=21,
	SHAPETYPE_POLYLINEM=23,
	SHAPETYPE_POLYGONM=25,
	SHAPETYPE_MULTIPOINTM=28,
	SHAPETYPE_MULTIPATCH=31
};

/* First 100 bytes of *.shp file */
#ifdef WIN32
#pragma pack(1)
#endif
struct shapefile_fileheader_struct {
	int32_t filecode_b;
	int32_t unused_b[5];
	int32_t filelength_b;	/* In 16-byte words */
	int32_t version_l;
	int32_t shapetype_l;
	double xmin;
	double ymin;
	double xmax;
	double ymax;
	double zmin;
	double zmax;
	double mmin;
	double mmax;
}
#ifndef WIN32
        __attribute__ ((packed))
#endif
;
#ifdef WIN32
#pragma pack()
#endif

/* First 100 bytes of *.shx file */
#ifdef WIN32
#pragma pack(1)
#endif
struct shapefile_indexfileheader_struct {
	int32_t filecode_b;
	int32_t unused_b[5];
	int32_t filelength_b;	/* In 16-byte words */
	int32_t version_l;
	int32_t shapetype_l;
	double xmin;
	double ymin;
	double xmax;
	double ymax;
	double zmin;
	double zmax;
	double mmin;
	double mmax;
}
#ifndef WIN32
        __attribute__ ((packed))
#endif
;
#ifdef WIN32
#pragma pack()
#endif

#ifdef WIN32
#pragma pack(1)
#endif
struct shapefile_point_struct {
	double x;
	double y;
}
#ifndef WIN32
        __attribute__ ((packed))
#endif
;
#ifdef WIN32
#pragma pack()
#endif

#ifdef WIN32
#pragma pack(1)
#endif
struct shapefile_pointm_struct {
	double x;
	double y;
	double m;
}
#ifndef WIN32
        __attribute__ ((packed))
#endif
;
#ifdef WIN32
#pragma pack()
#endif

#ifdef WIN32
#pragma pack(1)
#endif
struct shapefile_pointz_struct {
	double x;
	double y;
	double z;
	double m;
}
#ifndef WIN32
        __attribute__ ((packed))
#endif
;
#ifdef WIN32
#pragma pack()
#endif

#ifdef WIN32
#pragma pack(1)
#endif
struct shapefile_recordheader_struct {
	int32_t recordnumber_b;
	int32_t contentlength_b;
}
#ifndef WIN32
        __attribute__ ((packed))
#endif
;
#ifdef WIN32
#pragma pack()
#endif

#ifdef WIN32
#pragma pack(1)
#endif
struct shapefile_recordindex_struct {
	int32_t offset_b;
	int32_t contentlength_b;
}
#ifndef WIN32
        __attribute__ ((packed))
#endif
;
#ifdef WIN32
#pragma pack()
#endif

#ifdef WIN32
#pragma pack(1)
#endif
struct shapefile_record_struct {
	int32_t shapetype_l;
	union {
		struct {} null;

		struct shapefile_point_struct point;
		#ifdef WIN32
		#pragma pack(1)
		#endif
		struct {
			double box[4];
			int32_t numpoints;
			struct shapefile_point_struct *points;		/* array of size numpoints */
		} multipoint
		#ifndef WIN32
			__attribute__ ((packed))
		#endif
		;
		#ifdef WIN32
		#pragma pack()
		#endif

		#ifdef WIN32
		#pragma pack(1)
		#endif
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
		} polyline
		#ifndef WIN32
			__attribute__ ((packed))
		#endif
		;
		#ifdef WIN32
		#pragma pack()
		#endif

		#ifdef WIN32
		#pragma pack(1)
		#endif
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
		} polygon
		#ifndef WIN32
			__attribute__ ((packed))
		#endif
		;
		#ifdef WIN32
		#pragma pack()
		#endif

		struct shapefile_pointm_struct pointm;
		#ifdef WIN32
		#pragma pack(1)
		#endif
		struct {
			double box[4];
			int32_t numpoints;
			struct shapefile_pointm_struct *points;		/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} multipointm
		#ifndef WIN32
			__attribute__ ((packed))
		#endif
		;
		#ifdef WIN32
		#pragma pack()
		#endif

		#ifdef WIN32
		#pragma pack(1)
		#endif
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} polylinem
		#ifndef WIN32
			__attribute__ ((packed))
		#endif
		;
		#ifdef WIN32
		#pragma pack()
		#endif

		#ifdef WIN32
		#pragma pack(1)
		#endif
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} polygonm
		#ifndef WIN32
			__attribute__ ((packed))
		#endif
		;
		#ifdef WIN32
		#pragma pack()
		#endif

		struct shapefile_pointz_struct pointz;
		#ifdef WIN32
		#pragma pack(1)
		#endif
		struct {
			double box[4];
			int32_t numpoints;
			struct shapefile_pointm_struct *points;		/* array of size numpoints */
			double zmin;
			double zmax;
			double *zs;					/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} multipointz
		#ifndef WIN32
			__attribute__ ((packed))
		#endif
		;
		#ifdef WIN32
		#pragma pack()
		#endif

		#ifdef WIN32
		#pragma pack(1)
		#endif
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
			double zmin;
			double zmax;
			double *zs;					/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} polylinez
		#ifndef WIN32
			__attribute__ ((packed))
		#endif
		;
		#ifdef WIN32
		#pragma pack()
		#endif

		#ifdef WIN32
		#pragma pack(1)
		#endif
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
			double zmin;
			double zmax;
			double *zs;					/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} polygonz
		#ifndef WIN32
			__attribute__ ((packed))
		#endif
		;
		#ifdef WIN32
		#pragma pack()
		#endif

		#ifdef WIN32
		#pragma pack(1)
		#endif
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			int32_t *parttypes;				/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
			double zmin;
			double zmax;
			double *zs;					/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} multipatch
		#ifndef WIN32
			__attribute__ ((packed))
		#endif
		;
		#ifdef WIN32
		#pragma pack()
		#endif
	}
	#ifndef WIN32
		__attribute__ ((packed))
	#endif
	;
	#ifdef WIN32
	#pragma pack()
	#endif
}
#ifndef WIN32
        __attribute__ ((packed))
#endif
;
#ifdef WIN32
#pragma pack()
#endif

class ShapefilePolyLinePart {
public:
	ShapefilePolyLinePart();
	~ShapefilePolyLinePart();

	std::list<struct shapefile_point_struct*> *GetPoints();
	bool ContainsPoint(double Lat, double Lon);
	void Clear();
private:
	std::list<struct shapefile_point_struct*> Points;
};

class ShapefilePolyLine {
public:
	ShapefilePolyLine();
	~ShapefilePolyLine();

	std::vector<ShapefilePolyLinePart*> *GetParts();
	void Clear();
	void ResetBoundingBox();
	void SetBoundingBox(double LatMin, double LonMin, double LatMax, double LonMax);
	void GetBoundingBox(double *LatMin, double *LonMin, double *LatMax, double *LonMax);
	bool InBoundingBox(double Lat, double Lon);	/* Is a point in this shape's bounding box? */
	/* Is our bounding box partially or completely within the given one? */
	bool ContainedIn(double LatMin, double LonMin, double LatMax, double LonMax);
	bool ContainsPoint(double Lat, double Lon);
	int32_t GetnPoints();

private:
	std::vector<ShapefilePolyLinePart*> Parts;
	double BoundingBox[4];	/* LatMin, LonMin, LatMax, LonMax */
};

/* A Polygon is virtually identical to a PolyLine */
class ShapefilePolygon: public ShapefilePolyLine {
public:
private:
};

class Shapefile {
public:
	Shapefile();
	~Shapefile();

	std::vector<ShapefilePolygon*> *GetPolygons();
	std::vector<ShapefilePolyLine*> *GetPolyLines();
	void GetPolyLineBounds(double *LatMin,double  *LonMin,double  *LatMax,double  *LonMax);
	void Clear();
	void Load(std::string Filename);
	void LoadPolygons(FILE *fp);
	void LoadPolyLines(FILE *fp);
	void SavePolyLines(std::string Filename);

private:
	std::vector<ShapefilePolygon*> Polygons;
	std::vector<ShapefilePolyLine*> PolyLines;
};

#endif	/* SHAPEFILE_H_ */

